- Author: Xuanwo <github@xuanwo.io>
- Start Date: 2021-06-17
- RFC PR: [beyondstorage/specs#109](https://github.com/beyondstorage/specs/issues/109)
- Tracking Issue: [beyondstorage/go-storage#603](https://github.com/beyondstorage/go-storage/issues/603)

# GSP-109: Redesign Features

- Updated By:
    - [GSP-837: Support Feature Flag](./837-support-feature-flag.md): Reimplement `Features`

## Background

[GSP-87](https://github.com/beyondstorage/specs/pull/87) introduces a new concept `Feature`. However, it's proved to be too complicated that both hard to understand and to implemented.

In GSP-87, we added new struct `ServiceFeatures` and `StorageFeatures`:

```go
type ServiceFeatures struct {
   LooseOperationAll    bool
   LooseOperationCreate bool
   LooseOperationDelete bool
   LooseOperationGet    bool
   LooseOperationList   bool

   VirtualOperationAll bool

   VirtualPairAll bool
}

type StorageFeatures struct {
   LooseOperationAll               bool
   LooseOperationCompleteMultipart bool
   LooseOperationCreate            bool
   LooseOperationCreateMultipart   bool
   LooseOperationDelete            bool
   LooseOperationList              bool
   LooseOperationListMultipart     bool
   LooseOperationMetadata          bool
   LooseOperationRead              bool
   LooseOperationStat              bool
   LooseOperationWrite             bool
   LooseOperationWriteMultipart    bool

   VirtualOperationAll bool

   VirtualPairAll bool
}
```

There are the following problems:

- `LooseOperationXxx` is too verbose and there is no user who requires this feature.
- `VirtualOperationWriteDir` make developer-only care about it in `CreateDir`, but it's not correct: to implement `VirtualDir` support, the developer should check them everywhere.

## Proposal

Define features one by one and remove all autogenerated check, that means service implementer should check all features manually when required.

For now, we will introduce the following features, other features should be introduced by new GSPs.

### LoosePair

We will introduce a new feature called `LoosePair` to replace `LooseOperation` features.

If this feature is enabled, the service will not return an error for not support pairs.

### VirtualDir

We will introduce a new feature called `VirtualDir` to Replace `VirtualOperationCreateDir`.

Any operation that is related to dir SHOULD check `VirtualDir` in services that don't have native support for dir.

- `stat`: SHOULD check `VirtualDir` before set Dir ObjectMode
- `create`: SHOULD check `VirtualDir` before accept Dir ObjectMode
- `list`: SHOULD check `VirtualDir` before set Dir ObjectMode
- `delete`: SHOULD check `VirtualDir` before delete a dir Object.
- ...

### VirtualObjectMetadata

We will introduce a new feature called `VirtualObjectMetadata` to allow the service to handle object metadata even when it doesn't have native support.

This feature will replace `VirtualPairXxxYxxx` features like `VirutalPairWriteContentMd5` or `VirtualPairWriteContentType`. `VirtualPairXxxYxxx` used to added in GSP-87, but never implemented. In this proposal, we unify all of them into `VirtualObjectMetadata`.

## Rationale

N/A

## Compatibility

Related fields `ServiceFeatures` and `StorageFeatures` will be deprecated, like:

- `LooseOperationAll`
- `LooseOperationRead`
- `VirtualOperationAll`
- `VirtualOperationCreateDir`
- `VirtualPairWriteContentMd5`
- ...

The check logic will be removed but related fields will be marked as deprecated.

## Implementation

- Mark fields as deprecated
- Remove not used check logic
- Add features in specs and go-storage
- Implement feature checks in all service
